<?php

namespace InSegment\WorkersFramework\Drivers\RabbitMq;

use InSegment\WorkersFramework\Interfaces\ConsumerInterface;
use InSegment\WorkersFramework\Models\QueueConfig;
use Shavshukov\EventAppBase\Service\ErrorReport;

use Bunny\Protocol\MethodQueueDeclareOkFrame;
use Bunny\Protocol\MethodBasicConsumeOkFrame;
use Bunny\Async\Client;
use Bunny\Channel;
use Bunny\Message;
use React\Promise;

class AsyncChannel extends ChannelDriver
{
    /**
     * Create instance of RabbitMq service driver connected by async channel
     * 
     * @param \InSegment\WorkersFramework\Drivers\RabbitMq\Connector $connector
     * @param \React\Promise\PromiseInterface<\Bunny\Channel> $channelPromise
     * @return \React\Promise\PromiseInterface<\InSegment\WorkersFramework\Drivers\RabbitMq\AsyncChannel>
     */
    public static function create(Connector $connector, Promise\PromiseInterface $channelPromise): Promise\PromiseInterface
    {
        return $channelPromise->then(function (Channel $channel) use ($connector) {
            return $channel->qos(
                $connector->getPrefetchSize(),
                $connector->getPrefetchCount()    
            )->then(function () use ($connector, $channel) {
                return new AsyncChannel($connector, $channel);
            });
        });
    }
    
    /**
     * Open queue for work and get message count
     * 
     * @param \InSegment\WorkersFramework\Models\QueueConfig $configuration
     * @return \React\Promise\PromiseInterface<int>
     */
    public function openQueueForWork(QueueConfig $configuration)
    {
        return $this->declareWorkQueue($configuration)->then(function(MethodQueueDeclareOkFrame $result) {
            return $result->messageCount;
        });
    }
    
    /**
     * Declare queue described by configuration, and its pair queue for dlx, if necessary
     * 
     * @param \InSegment\WorkersFramework\Models\QueueConfig $configuration
     * @return \React\Promise\PromiseInterface
     */
    public function declareQueueStructure(QueueConfig $configuration)
    {
        return $this->channel->queueDelete($configuration->get('queueName'))->then(function () use ($configuration) {
            return $this->declareWorkQueue($configuration);
        })->then(function () use ($configuration) {
            return $this->channel->queueDelete($configuration->get('serviceQueueName'));
        })->then(function () use ($configuration) {
            return $this->declareServiceQueue($configuration);
        })->then(function () use ($configuration) {
            if (!$configuration->get('dlxQueueName')) {
                return Promise\resolve();
            }
            
            return $this->channel->queueDelete($configuration->get('dlxQueueName'))->then(function () use ($configuration) {
                return $this->declareDlxQueue($configuration);
            });
        });
    }

    /**
     * Subscribe to queue and return consumer tag
     * 
     * @param string $queueName
     * @param \InSegment\WorkersFramework\Interfaces\ConsumerInterface $receiver
     * @return \React\Promise\PromiseInterface<string>
     */
    public function subscribe(string $queueName, ConsumerInterface $receiver)
    {
        return $this->channel->consume(
            function (Message $message, Channel $channel, Client $client) use ($receiver, $queueName) {
                try {
                    $messageWrapped = new QueueMessage($message, $channel, $queueName);
                    $receiver->receiveMessage($messageWrapped, $this);
                } catch (\Throwable $e) {
                    ErrorReport::report($e, false);
                }
            },
            $queueName
        )->then(function (MethodBasicConsumeOkFrame $response) {
            return $response->consumerTag;
        });
    }

}
