<?php

namespace InSegment\WorkersFramework\Drivers\RabbitMq;

use InSegment\WorkersFramework\Interfaces\QueueMessageInterface;
use InSegment\WorkersFramework\Interfaces\QueueDriverInterface;
use InSegment\WorkersFramework\Models\QueueConfig;

use Bunny\Channel;

abstract class ChannelDriver implements QueueDriverInterface
{
    /**
     * RabbitMq connector
     *
     * @var \InSegment\WorkersFramework\Drivers\RabbitMq\Connector 
     */
    protected $connector;
    
    /**
     * RabbitMq channel
     *
     * @var \Bunny\Channel
     */
    protected $channel;
    
    /**
     * Constructor
     * 
     * @param \InSegment\WorkersFramework\Drivers\RabbitMq\Connector $connector
     * @param \Bunny\Channel $channel
     */
    public function __construct(Connector $connector, Channel $channel)
    {
        $this->connector = $connector;
        $this->channel = $channel;
    }
    
    /**
     * Publish message to the queue
     * 
     * @param \InSegment\WorkersFramework\Interfaces\QueueMessageInterface $message
     * @return bool|int|\React\Promise\PromiseInterface
     */
    public function publish(QueueMessageInterface $message)
    {
        return $this->channel->publish(
            $message->getContents(), // String
            $message->getHeaders()->all(), // Headers
            '', // Exchange name
            $message->getQueueName() // Routing key
        );
    }
    
    /**
     * Un-subscribe from channel
     * 
     * @param string $consumerTag
     * @return bool|\Bunny\Protocol\MethodBasicCancelOkFrame|\React\Promise\PromiseInterface<\Bunny\Protocol\MethodBasicCancelOkFrame>
     */
    public function unsubscribe(string $consumerTag)
    {
        return $this->channel->cancel($consumerTag);
    }
    
    /**
     * Get message count from queue
     * 
     * @param \InSegment\WorkersFramework\Models\QueueConfig $configuration
     * @return int|\React\Promise\PromiseInterface<int>
     */
    public function getMessageCount(QueueConfig $configuration)
    {
        return $this->openQueueForWork($configuration);
    }
    
    /**
     * Declare work queue for queue described by configuration
     * 
     * @param \InSegment\WorkersFramework\Models\QueueConfig $configuration
     * @return boolean|Promise\PromiseInterface|Protocol\MethodQueueDeclareOkFrame
     */
    protected function declareWorkQueue(QueueConfig $configuration)
    {
        $arguments = [];

        if ($configuration->get('dlxQueueName')) {
            $arguments['x-dead-letter-exchange'] = '';
            $arguments['x-dead-letter-routing-key'] = $configuration->get('dlxQueueName');
        }

        return $this->channel->queueDeclare($configuration->get('queueName'), false, true, false, false, false, $arguments);
    }
    
    /**
     * Declare service queue for queue described by configuration
     * 
     * @param \InSegment\WorkersFramework\Models\QueueConfig $configuration
     * @return boolean|Promise\PromiseInterface|Protocol\MethodQueueDeclareOkFrame
     */
    protected function declareServiceQueue(QueueConfig $configuration)
    {
        return $this->channel->queueDeclare($configuration->get('serviceQueueName'));
    }
    
    /**
     * Declare dlx queue for queue described by configuration
     * 
     * @param \InSegment\WorkersFramework\Models\QueueConfig $configuration
     * @return boolean|Promise\PromiseInterface|Protocol\MethodQueueDeclareOkFrame
     */
    protected function declareDlxQueue(QueueConfig $configuration)
    {
        $arguments = [];

        if ($configuration->get('dlxRequeue')) {
            if (($timeout = $configuration->get('dlxTimeout'))) {
                $arguments['x-message-ttl'] = $timeout;
            }

            $arguments['x-dead-letter-exchange'] = '';
            $arguments['x-dead-letter-routing-key'] = $configuration->get('queueName');
        }

        return $this->channel->queueDeclare($configuration->get('dlxQueueName'), false, true, false, false, false, $arguments);
    }
}
