<?php

namespace InSegment\WorkersFramework\Drivers\RabbitMq;

use Shavshukov\EventAppBase\ApplicationInterface;
use Shavshukov\EventAppBase\Service\ErrorReport;

use Bunny\Async\Client as AsyncClient;
use Bunny\Client;
use Bunny\Channel;
use React\Promise;

class Connector
{
    /**
     * Memory queue channel
     *
     * @var \Bunny\Channel
     */
    protected $channel;
    
    /**
     * Application instance
     *
     * @var \Shavshukov\EventAppBase\ApplicationInterface 
     */
    protected $app;
    
    /**
     * Connection options
     * 
     * @var array
     */
    protected $options;
    
    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\ApplicationInterface $app
     */
    public function __construct(ApplicationInterface $app)
    {
        $this->app = $app;
        $this->setOptions();
    }
    
    /**
     * Get prefetch size from application config
     * 
     * @return int
     */
    public function getPrefetchSize(): int
    {
        return $this->app->config('mq_prefetch_size', 0);
    }
    
    /**
     * Get prefetch count from application config
     * 
     * @return int
     */
    public function getPrefetchCount(): int
    {
        return $this->app->config('mq_prefetch_count', 1);
    }
    
    /**
     * Create sync or async driver
     * 
     * @param bool $sync
     * @return \InSegment\WorkersFramework\Drivers\RabbitMq\QueueDriverInterface|\React\Promise\PromiseInterface<\InSegment\WorkersFramework\Drivers\RabbitMq\QueueDriverInterface>
     */
    public function createConnectedChannelDriver(bool $sync = false)
    {
        if ($sync) {
            return SyncChannel::create($this, $this->createSyncChannel());
        }
        
        return AsyncChannel::create($this, $this->createAsyncChannel());
    }
    
    /**
     * Create sync channel driver
     * 
     * @return \Bunny\Channel
     */
    protected function createSyncChannel(): Channel
    {
        $bunny = new Client($this->options);
        $bunny->connect();
        return $bunny->channel();
    }
    
    /**
     * Create async channel driver
     * 
     * @return \React\Promise\PromiseInterface<\Bunny\Channel>
     */
    protected function createAsyncChannel(): Promise\PromiseInterface
    {
        return (new AsyncClient($this->app->getLoop(), $this->options))->connect()->then(function (AsyncClient $client) {
            return $client->channel();
        }, function (\Exception $e) {
            ErrorReport::report($e, true);
            exit(-1);
        });
    }
    
    /**
     * Get options for mq
     * 
     * @return null
     */
    protected function setOptions()
    {
        $config = $this->app->config();
        $this->options = [
            'host' => $config->get('mq_host'),
            'vhost' => $config->get('mq_vhost'),
            'user' => $config->get('mq_user'),
            'password' => $config->get('mq_password'),
        ];
    }
}
