<?php

namespace InSegment\WorkersFramework\Drivers\RabbitMq;

use InSegment\WorkersFramework\Models\QueueConfig;
use Shavshukov\EventAppBase\Service\ErrorReport;

use Bunny\Client;
use Bunny\Channel;
use Bunny\Message;

class SyncChannel extends ChannelDriver
{
    
    /**
     * Create instance of RabbitMq service driver connected by async channel
     * 
     * @param \InSegment\WorkersFramework\Drivers\RabbitMq\Connector $connector
     * @param \Bunny\Channel $channel
     * @return \InSegment\WorkersFramework\Drivers\RabbitMq\SyncChannel
     */
    public static function create(Connector $connector, Channel $channel)
    {
        return new AsyncChannel($connector, $channel);
    }
    
    /**
     * Open queue for work and get message count
     * 
     * @param \InSegment\WorkersFramework\Models\QueueConfig $configuration
     * @return int
     */
    public function openQueueForWork(QueueConfig $configuration)
    {
        return $this->declareWorkQueue($configuration)->messageCount;
    }

    /**
     * Declare queue described by configuration, and its pair queue for dlx, if necessary
     * 
     * @param \InSegment\WorkersFramework\Models\QueueConfig $configuration
     * @return null
     */
    public function declareQueueStructure(QueueConfig $configuration)
    {
        $configValidation = $configuration->validate();
        
        if (!$configValidation->isValid()) {
            $printNotFoundFields = implode(', ', $configValidation->getInvalidFields());
            throw new Exception("Queue configuration mandatory fields not found: {$printNotFoundFields}");
        }
        
        $this->channel->queueDelete($configuration->get('queueName'));
        $this->declareWorkQueue($configuration);
        $this->channel->queueDelete($configuration->get('serviceQueueName'));
        $this->declareServiceQueue($configuration);
        
        if ($configuration->dlxQueueName) {
            $this->channel->queueDelete($configuration->get('dlxQueueName'));
            $this->declareDlxQueue($configuration);
        }
    }

    /**
     * Subscribe to queue and return consumer tag
     * 
     * @param string $queueName
     * @param \InSegment\WorkersFramework\Interfaces\ConsumerInterface $receiver
     * @return string
     */
    public function subscribe(string $queueName, ConsumerInterface $receiver)
    {
        $response = $this->channel->consume(
            function (Message $message, Channel $channel, Client $client) use ($receiver, $queueName) {
                try {
                    $messageWrapped = new QueueMessage($message, $channel, $queueName);
                    $receiver->receiveMessage($messageWrapped, $this);
                } catch (\Throwable $e) {
                    ErrorReport::report($e, false);
                }
            },
            $queueName
        );
            
        return $response->consumerTag;
    }
    
}
