<?php

namespace InSegment\WorkersFramework\Models;

use Shavshukov\EventAppBase\FileReaders\FileReaderInterface;
use Shavshukov\EventAppBase\Model\ModelEventsTrait;
use Shavshukov\EventAppBase\Model\YamlModel;
use Shavshukov\EventAppBase\Storage\FileStorage;

use InSegment\WorkersFramework\Interfaces\JobInterface;

class Config extends YamlModel
{
    use ModelEventsTrait;
    
    const DEFAULT_CONFIG_FILE  = 'config.yaml';
    const CONFIG_MANDATORY_FIELDS = ['main_queue', 'service_queue', 'worker_run_command'];
    const WORKER_MANDATORY_FIELDS = ['work_queue', 'service_queue'];
    
    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\FileReaders\FileReaderInterface $fileReader
     * @param string $fileName
     */
    public function __construct(FileReaderInterface $fileReader, string $fileName)
    {
        $fileStorage = new FileStorage($fileReader, $fileName);
        $this->backend = $fileStorage;
        $this->bootEventsIfNotBooted();
    }
    
    /**
     * Get an instance of JobInterface by the name of the job
     * 
     * @param string $jobName
     * @return \InSegment\WorkersFramework\Interfaces\JobInterface|null
     */
    public function getJob(string $jobName): ?JobInterface
    {
        $jobClass = $this->get("jobs.{$jobName}");
        if ($jobClass) {
            return new $jobClass();
        }
    }
    
    /**
     * Register events once on the class
     * 
     * @return null
     */
    protected static function bootEvents()
    {
        static::loaded(function ($model, $result) {
            if (!$result) {
                throw new \Exception('Unable to load config file');
            }
            
            if (!is_array($model->data['workers'])) {
                throw new \Exception('Config checking: Not found Workers array');
            }

            if (!is_array($model->data['jobs'])) {
                throw new \Exception('Config checking: Not found Jobs array');
            }

            if (($notFoundFields = array_diff(self::CONFIG_MANDATORY_FIELDS, array_keys($model->data)))) {
                $printNotFoundFields = implode(', ', $notFoundFields);
                throw new \Exception("Config checking: Fields not found: {$printNotFoundFields}");
            }
            
            foreach ($model->data['workers'] as $workerKey => $workerConfig) {
                if (($notFoundFields = array_diff(self::WORKER_MANDATORY_FIELDS, array_keys($workerConfig)))) {
                    $printNotFoundFields = implode(', ', $notFoundFields);
                    throw new \Exception("Workers config checking: Fields not found: {$printNotFoundFields} for key {$workerKey}");
                }
            }
            
            foreach ($model->data['jobs'] as $jobName => $jobClass) {
                if (!class_exists($jobClass)) {
                    throw new \Exception("Jobs config checking: Class not found: {$jobClass} for job {$jobName}");
                }
                
                try {
                    if (!(new $jobClass) instanceof JobInterface) {
                        throw new \Exception("Jobs config checking: {$jobClass} is not an implementor of JobInterface!");
                    }
                } catch (\Throwable $e) {
                    throw new \Exception("Cannot create an instance of {$jobClass}!");
                }
            }

            return true;
        });
    }
}
