<?php

namespace InSegment\WorkersFramework\Queues;

use Shavshukov\EventAppBase\Model\ArrayModel;
use InSegment\WorkersFramework\Models\QueueConfig;

class Worker
{
    /**
     * Name of the worker
     *
     * @var string
     */
    protected $workerName;
    
    /**
     * Name of the queue this worker consumes
     * 
     * @var string
     */
    protected $queueName;
    
    /**
     * Name of the queue for state update events for this worker
     * 
     * @var string
     */
    protected $serviceQueueName;
    
    /**
     * Worker will prefer jobs that he attracts
     *
     * @var array
     */
    protected $attracts;
    
    /**
     * Worker will not prefer jobs that ha avoids
     *
     * @var array
     */
    protected $avoids;
    
    /**
     * Status of the worker
     *
     * @var \Shavshukov\EventAppBase\Model\ArrayModel 
     */
    protected $status;
    
    /**
     * Configuration model
     *
     * @var \InSegment\WorkersFramework\Models\QueueConfig 
     */
    protected $queueConfig;
    
    /**
     * Constructor
     * 
     * @param string $workerName
     * @param array $config
     */
    public function __construct(string $workerName, array $config)
    {
        $this->workerName = $workerName;
        $this->queueName = $config['work_queue'];
        $this->serviceQueueName = $config['service_queue'];
        $this->attracts = $config['attracts'] ?? [];
        $this->avoids = $config['avoids'] ?? [];
        $this->jobWeight = $config['job_weight'] ?? 1;
        $this->status = ArrayModel::from([
            'active' => false,
            'workerState' => 'offline',
            'startTime' => null
        ]);
        $this->queueConfig = new QueueConfig([
            'queueName' => $config['work_queue'],
            'serviceQueueName' => $config['service_queue'],
            'dlxQueueName' => $config['dlx_queue'] ?? null,
            'dlxRequeue' => true,
            'dlxTimeout' => $config['dlx_timeout'] ?? null
        ]);
        $this->queueConfig->check();
    }
    
    /**
     * Get Consumer status as array
     * 
     * @return array
     */
    public function getStatus(): array
    {
        return [
            'worker_state' => $this->status->get('workerState'),
            'active' => $this->status->get('active'),
        ];
    }
    
    /**
     * Whether there is currently a process under that worker
     * 
     * @return bool
     */
    public function isActive(): bool
    {
        return $this->status->get('active');
    }
    
    /**
     * Change values in status of the worker according to updates from the process
     * 
     * @param array $modification
     * @return null
     */
    public function modifyStatus(array $modification)
    {
        $this->status->modify($modification);
    }
    
    /**
     * Get name of the worker
     * 
     * @return string
     */
    public function getName(): string
    {
        return $this->workerName;
    }
    
    /**
     * Get name of the queue this worker consumes
     * 
     * @return string
     */
    public function getQueueName(): string
    {
        return $this->queueName;
    }
    
    /**
     * Get name of the queue for state update events for this worker
     * 
     * @return string
     */
    public function getServiceQueueName(): string
    {
        return $this->serviceQueueName;
    }
    
    /**
     * Get jobs that this worker attracts
     * 
     * @return array
     */
    public function getAttracts(): array
    {
        return $this->attracts;
    }
    
    /**
     * Get jobs that this worker avoids
     * 
     * @return array
     */
    public function getAvoids(): array
    {
        return $this->avoids;
    }
    
    /**
     * Get queue config of the worker
     * 
     * @return \InSegment\WorkersFramework\Queues\QueueConfig
     */
    public function getQueueConfig(): QueueConfig
    {
        return $this->queueConfig;
    }
}

