<?php

namespace InSegment\WorkersFramework\Services;

use InSegment\WorkersFramework\Interfaces\WorkerListInterface;
use InSegment\WorkersFramework\Interfaces\JobInterface;
use InSegment\WorkersFramework\Queues\Worker;

class Distributor
{
    const DEFAULT_SCORE = 0;
    const MIN_SCORE = -20;
    const MAX_SCORE = 20;
    const ATTRACTS_INCREASE = self::MIN_SCORE;
    const AVOIDS_DECREASE = self::MAX_SCORE;
    
    const COMPARISON_EQUAL = 0;
    const COMPARISON_LESS = -1;
    const COMPARISON_GREATER = 1;
    
    /**
     * List of workers to choose from
     *
     * @var \InSegment\WorkersFramework\Interfaces\WorkerListInterface
     */
    protected $workersList;
    
    /**
     * Load measure instance
     *
     * @var \InSegment\WorkersFramework\Queues\LoadMeasure
     */
    protected $loadMeasure;
    
    /**
     * Constructor
     * 
     * @param \InSegment\WorkersFramework\Interfaces\WorkerListInterface $workersList
     */
    public function __construct(WorkerListInterface $workersList)
    {
        $this->workersList = $workersList;
        $this->loadMeasure = $workersList->getLoadMeasure();
    }
    
    /**
     * Find the most suitable worker to do the job
     * 
     * @param JobInterface $job
     * @return \InSegment\WorkersFramework\Services\Worker|null
     */
    public function findBestWorkerFor(JobInterface $job): ?Worker
    {
        $name = $job->getName();
        $workers = $this->workersList->getWorkers();
        $workerScores = [];

        if (empty($workers)) {
            echo "Warning: no workers available!\n";
            return null;
        } else if (count($workers) === 1) {
            echo "Only one worker available: {$workers[0]->getName()}\n";
            return $workers[0];
        }
        
        usort($workers, function (Worker $worker1, Worker $worker2) use ($name, &$workerScores) {
            // scores will define the order of workers
            $score1 = $score2 = self::DEFAULT_SCORE;
            
            // attracts works like 'recommendation whitelist', it is as white, but not as strict
            $attracts1 = $worker1->getAttracts();
            $attracts2 = $worker2->getAttracts();
            $isAttracts1 = !empty($attracts1[$name]);
            $isAttracts2 = !empty($attracts2[$name]);
            
            // prefer workers, who attract the job
            if ($isAttracts1) { $score1 += self::ATTRACTS_INCREASE; }
            if ($isAttracts2) { $score2 += self::ATTRACTS_INCREASE; }
            
            // avoids works like 'recommendation blacklist', it is as black, but not as strict
            $avoids1 = $worker1->getAttracts();
            $avoids2 = $worker2->getAttracts();
            $isAvoids1 = !$isAttracts1 && (!empty($avoids1[$name]) || $avoids1 === ['*']);
            $isAvoids2 = !$isAttracts2 && (!empty($avoids2[$name]) || $avoids2 === ['*']);
            
            // prefer workers, who doesn't avoid the job
            if ($isAvoids1) { $score1 += self::AVOIDS_DECREASE; }
            if ($isAvoids2) { $score2 += self::AVOIDS_DECREASE; }
            
            // prefer a worker that is less loaded
            $load1 = $this->loadMeasure->getLoad($worker1->getName());
            $load2 = $this->loadMeasure->getLoad($worker2->getName());
            
            $score1 += $load1;
            $score2 += $load2;
            
            // normalize scores
            if ($score1 < self::MIN_SCORE) {
                $score1 = self::MIN_SCORE;
            } else if ($score1 > self::MAX_SCORE) {
                $score1 = self::MAX_SCORE;
            }
            
            if ($score2 < self::MIN_SCORE) {
                $score2 = self::MIN_SCORE;
            } else if ($score2 > self::MAX_SCORE) {
                $score2 = self::MAX_SCORE;
            }
            
            $workerScores[$worker1->getName()] = $score1;
            $workerScores[$worker2->getName()] = $score2;
            
            if ($score1 === $score2) {
                return self::COMPARISON_EQUAL;
            }
            
            // prefer a worker that scored less
            return $score1 < $score2 ? self::COMPARISON_LESS : self::COMPARISON_GREATER;
        });
        
        $selectedWorker = $workers[0];
        $selectedWorkerName = $selectedWorker->getName();
        $competitors = [];
        foreach ($workerScores as $workerName => $score) {
            $competitors[] = "{$workerName}:{$score}";
        }
        $printCompetitors = '['. implode(', ', $competitors) . ']';
        
        echo "Distributor:\n\tselected worker: {$selectedWorkerName}\n\tfor job {$name}\n\twith score: {$workerScores[$selectedWorkerName]}\n\tCompetitors: {$printCompetitors}\n";
        return $selectedWorker;
    }
}
