<?php

namespace InSegment\WorkersFramework\Services;

use React\Promise;

use InSegment\WorkersFramework\Queues\UnpublishedMessage;
use InSegment\WorkersFramework\Interfaces\QueueMessageInterface;
use InSegment\WorkersFramework\Interfaces\QueueDriverInterface;
use InSegment\WorkersFramework\Interfaces\ConsumerInterface;
use InSegment\WorkersFramework\Traits\EventsTrait;
use InSegment\WorkersFramework\Models\JobHeaders;
use InSegment\WorkersFramework\Models\Config;

class MainConsumer implements ConsumerInterface
{
    use EventsTrait;
    
    /**
     * Instance of queue driver connected to the channel
     *
     * @var \InSegment\WorkersFramework\Interfaces\QueueDriverInterface
     */
    protected $queueDriver;
    
    /**
     * Distributor instance
     *
     * @var \InSegment\WorkersFramework\Services\Distributor
     */
    protected $distrubutor;
    
    /**
     * Config instance
     *
     * @var \InSegment\WorkersFramework\Models\Config
     */
    protected $config;
    
    /**
     * Whether the consumer is started
     *
     * @var bool
     */
    protected $active;
    
    /**
     * Tag to track active subscription
     *
     * @var string|null
     */
    protected $consumerTag;
    
    /**
     * Name of the main queue
     *
     * @var string
     */
    protected $queueName;
    
    /**
     * Constructor
     * 
     * @param \InSegment\WorkersFramework\Interfaces\QueueDriverInterface $driver
     * @param \InSegment\WorkersFramework\Services\Distributor $distributor
     * @param \InSegment\WorkersFramework\Models\Config $config
     */
    public function __construct(QueueDriverInterface $driver, Distributor $distributor, Config $config)
    {
        $this->queueDriver = $driver;
        $this->distrubutor = $distributor;
        $this->config = $config;
        $this->queueName = $config->get('main_queue');
    }
    
    /** 
     * Activate consumption of the main queue
     * 
     * @return \React\Promise\PromiseInterface
     */
    public function activate(): Promise\PromiseInterface
    {
        echo "Distribution consumer starting: {$this->queueName}\n";
        
        if ($this->active) { //  already started
            echo "already started: {$this->queueName}\n";
            return Promise\resolve();
        }

        return $this->queueDriver->subscribe($this->queueName, $this)->then(function ($consumerTag) {
            $this->active = true;
            $this->consumerTag = $consumerTag;
            $this->trigger('started', $consumerTag);
            echo "Distribution consume started: {$this->queueName}\n";
        });
    }
    
    /**
     * Deactivate consumption of the main queue
     * 
     * @return \React\Promise\PromiseInterface
     */
    public function deactivate(): Promise\PromiseInterface
    {
        echo "Distribution consumer stopping: {$this->queueName}\n";
        
        if (!$this->active) {
            echo "not started: {$this->queueName}\n";
            return Promise\resolve();
        }
        
        $consumerTag = $this->consumerTag;
        return $this->queueDriver->unsubscribe($consumerTag)->then(function () use ($consumerTag) {
            $this->trigger('stopped', $consumerTag);
            $this->active = false;
        });
    }
    
    /**
     * Move message from the main queue to the worker queue and acknowledge the original one
     * 
     * @param \InSegment\WorkersFramework\Interfaces\QueueMessageInterface $message
     * @param \InSegment\WorkersFramework\Interfaces\QueueDriverInterface $driver
     * @return mixed
     */
    public function receiveMessage(QueueMessageInterface $message, QueueDriverInterface $driver)
    {
        $headers = JobHeaders::extend($message->getHeaders());
        $headersValidation = $headers->validate();
        
        if (!$headersValidation->isValid()) {
            $printNotFoundFields = implode(', ', $headersValidation->getInvalidFields());
            echo "Job header mandatory fields not found: {$printNotFoundFields}, message will be ignored\n";
            return $message->acknowlege();
        }
        
        $jobName = $headers->getJobName();
        $job = $this->config->getJob($jobName);
        
        if ($job === null) {
            echo "The configuration doesn't have class mapping for job named {$jobName}, message will be ignored\n";
            return $message->acknowlege();
        }
        
        $worker = $this->distrubutor->findBestWorkerFor($job);
        $processingQueueName = $worker->getQueueName();
        
        echo "Distributing message with job {$jobName} to queue {$worker->getQueueName()}\n";
        $newMessage = new UnpublishedMessage($processingQueueName, $message->getContents(), $message->getHeaders());
        $this->trigger('publishing', $worker, $job, $newMessage);
        return $driver->publish($newMessage)->then(function() use ($message, $worker, $job, $newMessage) {
            $this->trigger('published', $worker, $job, $newMessage);
            return $message->acknowlege();
        });
    }

}
