<?php

namespace InSegment\WorkersFramework\Services;

use Shavshukov\EventAppBase\ApplicationInterface;
use Shavshukov\EventAppBase\Interfaces\ModelInterface;
use React\Promise;

use InSegment\WorkersFramework\Interfaces\QueueDriverInterface;
use InSegment\WorkersFramework\Interfaces\QueueMessageInterface;
use InSegment\WorkersFramework\Interfaces\WorkerListInterface;
use InSegment\WorkersFramework\Interfaces\JobInterface;
use InSegment\WorkersFramework\Services\MainConsumer;
use InSegment\WorkersFramework\Http\Server;
use InSegment\WorkersFramework\Models\State;
use InSegment\WorkersFramework\Queues\Worker;
use InSegment\WorkersFramework\Queues\LoadMeasure;
use InSegment\WorkersFramework\Queues\ServiceMessenger;

class Manager implements WorkerListInterface
{
    /**
     * Application instance
     *
     * @var \Shavshukov\EventAppBase\ApplicationInterface
     */
    protected $app;
    
    /**
     * Instance of queue driver connected to the channel
     *
     * @var \InSegment\WorkersFramework\Interfaces\QueueDriverInterface
     */
    protected $queueDriver;
    
    /**
     * HTTP server instance
     *
     * @var \InSegment\WorkersFramework\Http\Server
     */
    protected $httpServer;
    
    /**
     * Configuration of each worker
     *
     * @var array
     */
    protected $workersConfig;
    
    /**
     * Consumer for main queue
     *
     * @var \InSegment\WorkersFramework\Services\MainConsumer
     */
    protected $mainConsumer;
    
    /**
     * Consumer for state queue
     *
     * @var \InSegment\WorkersFramework\Services\StateConsumer
     */
    protected $stateConsumer;
    
    /**
     * State instance
     *
     * @var \InSegment\WorkersFramework\Models\State
     */
    protected $state;
    
    /**
     * All workers initialised
     *
     * @var \InSegment\WorkersFramework\Queues\Worker[]
     */
    protected $workers;
    
    /**
     * Worker processes
     *
     * @var \InSegment\WorkersFramework\Services\WorkerProcess[]
     */
    protected $workerProcesses;
    
    /**
     * Messengers for workers service processes
     *
     * @var \InSegment\WorkersFramework\Queues\ServiceMessenger[]
     */
    protected $workerMessengers;
    
    /**
     * Load measure instance
     *
     * @var \InSegment\WorkersFramework\Queues\LoadMeasure
     */
    protected $loadMeasure;
    
    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\ApplicationInterface $app
     * @param \InSegment\WorkersFramework\Interfaces\QueueDriverInterface $queueDriver
     * @throws \Exception
     */
    public function __construct(ApplicationInterface $app, QueueDriverInterface $queueDriver)
    {
        $this->app = $app;
        $this->queueDriver = $queueDriver;
        $this->httpServer = new Server($app);
        $this->workersConfig = $app->config('workers', []);
        $this->state = new State($this->app->getFileReader(), $this->app->getAppPath());
        $this->loadMeasure = new LoadMeasure(array_keys($this->workersConfig));
        $this->mainConsumer = new MainConsumer($queueDriver, new Distributor($this), $app->config());
        $this->mainConsumer->on('started', function() {
            $modification = ['locked' => false];
            $this->state->modify($modification)->save();
            
            echo "Starting worker processes...\n";
            $enabledWorkers = $this->getWorkers();
            foreach ($enabledWorkers as $enabledWorker) {
                if (!$enabledWorker->isActive()) {
                    $this->workerProcesses[$enabledWorker->getName()]->start();
                }
            }
        });
        $this->mainConsumer->on('published', function (Worker $worker, JobInterface $job, QueueMessageInterface $newMessage) {
            $this->loadMeasure->increment($worker->getName(), max([$job->getWeight(), 1]));
        });
        $this->mainConsumer->on('stopped', function() {
            $modification = ['locked' => true];
            $this->state->modify($modification)->save();
            
            echo "Sending shutdown updates to worker service queues...\n";
            $promises = [];
            foreach ($this->workers as $workerName => $worker) {
                if ($worker->isActive()) {
                    $promises[] = $this->workerMessengers[$workerName]->sendUpdates($modification);
                }
            }
            
            Promise\all($promises)->then(function() {
                echo "Updates to workers were sent successfully...\n";
            });
        });
        $this->stateConsumer = new StateConsumer($queueDriver, $this->state, $app->config('service_queue'));
        $this->stateConsumer->on('updates', function (ModelInterface $updates) {
            if ($updates->get('locked')) {
                $this->deactivateConsuming();
            }
            
            $this->state->save();
        });
        $this->initWorkers();
        $app->service()->bind('manager', $this);
    }

    /**
     * Init worker consumers and Http server
     * 
     * @return \React\Promise\PromiseInterface
     */
    public function manage(): Promise\PromiseInterface
    {
        $promises = [];
        $this->state->load();
        $this->state->modify(['timestamp' => microtime(true)]);
        $promises[] = $this->stateConsumer->activate();
        
        if (!$this->state->get('locked')) {
            $promises[] = $this->activateConsuming();
        }
        
        echo "Starting http server...\n";
        $promises[] = $this->httpServer->serve();
        
        return Promise\all($promises);
    }

    /**
     * Activate consuming of the main queue
     * 
     * @return \React\Promise\PromiseInterface
     */
    public function activateConsuming(): Promise\PromiseInterface
    {
        echo "Activating consume of the main queue...\n";
        return $this->mainConsumer->activate();
    }
    
    /**
     * Activate particular worker
     * 
     * @param string $workerName
     * @return \React\Promise\PromiseInterface
     * @throws \Exception
     */
    public function activateWorker(string $workerName): Promise\PromiseInterface
    {
        if (!isset($this->workers[$workerName])) {
            throw new \Exception("Worker {$workerName} does not exists");
        }
        
        $worker = $this->workers[$workerName];
        $this->loadMeasure->add($workerName);
        
        $modification = [
            'workers' => [
                $workerName => [
                    'locked' => false
                ]
            ]
        ];
        $this->state->modify($modification)->save();
        
        if (!($this->state->get('locked') || $worker->isActive())) {
            $this->workerProcesses[$workerName]->start();
        }
        
        return Promise\resolve();
    }

    /**
     * Deactivate consuming of the main queue
     * 
     * @return \React\Promise\PromiseInterface
     */
    public function deactivateConsuming(): Promise\PromiseInterface
    {
        echo "Deactivating consume of the main queue...\n";
        foreach ($this->workers as $worker) {
            if ($worker->isActive()) {
                $worker->modifyStatus(['workerState' => 'stopping...']);
            }
        }
        
        return $this->mainConsumer->deactivate();
    }
    
    /**
     * Deactivate particular worker
     * 
     * @param string $workerName
     * @return \React\Promise\PromiseInterface
     * @throws \Exception
     */
    public function deactivateWorker(string $workerName): Promise\PromiseInterface
    {
        if (!isset($this->workers[$workerName])) {
            throw new \Exception("Worker {$workerName} does not exists");
        }
        
        $worker = $this->workers[$workerName];
        $this->loadMeasure->remove($workerName);
        
        $modification = [
            'workers' => [
                $workerName => [
                    'locked' => true
                ]
            ]
        ];
        $this->state->modify($modification)->save();
        
        if (!$this->state->get('locked') && $worker->isActive()) {
            echo "Sending shutdown update to worker {$workerName}...\n";
            return $this->workerMessengers[$workerName]->sendUpdates($modification);
        }
        
        return Promise\resolve();
    }
    
    /**
     * Get queue status as array
     * 
     * @return array
     */
    public function getStatus(): array
    {
        $workerStatus = [];
        foreach ($this->workers as $worker) {
            $workerStatus[$worker->getName()] = $worker->getStatus();
        }

        return [
            'workers_config' => $this->workersConfig,
            'consumers' => $workerStatus,
            'state' => $this->state->all(),
        ];
    }

    /**
     * Get available workers
     * 
     * @return \InSegment\WorkersFramework\Services\Worker[]
     */
    public function getWorkers(): array
    {
        $availableWorkers = [];
        foreach ($this->workers as $worker) {
            if (!$this->state->isWorkerLocked($worker->getName())) {
                $availableWorkers[] = $worker;
            }
        }
        
        return $availableWorkers;
    }

    /**
     * Get load measure instance
     * 
     * @return \InSegment\WorkersFramework\Queues\LoadMeasure
     */
    public function getLoadMeasure(): LoadMeasure
    {
        return $this->loadMeasure;
    }
    
    /**
     * Prepare all workers and their processes
     * 
     * @return null
     */
    protected function initWorkers()
    {
        $workers = [];
        $workerProcesses = [];
        $workerMessengers = [];
        
        foreach ($this->workersConfig as $workerName => $config) {
            $worker = new Worker($workerName, $config);
            $workerProcesses[$workerName] = new WorkerProcess($this->app, $this->state, $worker);
            $workerMessengers[$workerName] = new ServiceMessenger($this->queueDriver, $worker->getServiceQueueName());
            $workers[$workerName] = $worker;
        }
        
        $this->workers = $workers;
        $this->workerProcesses = $workerProcesses;
        $this->workerMessengers = $workerMessengers;
    }
}
