<?php

namespace InSegment\WorkersFramework\Services;

use Shavshukov\EventAppBase\Model\JsonModel;
use React\Promise;

use InSegment\WorkersFramework\Interfaces\QueueMessageInterface;
use InSegment\WorkersFramework\Interfaces\QueueDriverInterface;
use InSegment\WorkersFramework\Interfaces\ConsumerInterface;
use InSegment\WorkersFramework\Traits\EventsTrait;
use InSegment\WorkersFramework\Models\State;

class StateConsumer implements ConsumerInterface
{
    use EventsTrait;
    
    /**
     * Instance of queue driver connected to the channel
     *
     * @var \InSegment\WorkersFramework\Interfaces\QueueDriverInterface
     */
    protected $queueDriver;
    
    /**
     * State instance
     *
     * @var \InSegment\WorkersFramework\Models\State
     */
    protected $state;
    
    /**
     * Whether the consumer is started
     *
     * @var bool
     */
    protected $active;
    
    /**
     * Tag to track active subscription
     *
     * @var string|null
     */
    protected $consumerTag;
    
    /**
     * Name of the service queue
     *
     * @var string
     */
    protected $queueName;
    
    /**
     * Constructor
     * 
     * @param \InSegment\WorkersFramework\Interfaces\QueueDriverInterface $driver
     * @param \InSegment\WorkersFramework\Models\State $state
     * @param string $queueName
     */
    public function __construct(QueueDriverInterface $driver, State $state, string $queueName)
    {
        $this->queueDriver = $driver;
        $this->state = $state;
        $this->queueName = $queueName;
    }
    
    /**
     * Activate consumption of the main queue
     * 
     * @return \React\Promise\PromiseInterface
     */
    public function activate(): Promise\PromiseInterface
    {
        echo "State sync consumer starting: {$this->queueName}\n";
        
        if ($this->active) { //  already started
            echo "already started: {$this->queueName}\n";
            return Promise\resolve();
        }

        return $this->queueDriver->subscribe($this->queueName, $this)->then(function ($consumerTag) {
            $this->active = true;
            $this->consumerTag = $consumerTag;
            echo "State sync consume started: {$this->queueName}\n";
        });
    }
    
    /**
     * Move message from the main queue to the worker queue and acknowledge the original one
     * 
     * @param \InSegment\WorkersFramework\Interfaces\QueueMessageInterface $message
     * @param \InSegment\WorkersFramework\Interfaces\QueueDriverInterface $driver
     * @return mixed
     */
    public function receiveMessage(QueueMessageInterface $message, QueueDriverInterface $driver)
    {
        $updates = JsonModel::from($message->getContents());
        $microtimeUpdate = (float) $updates->get('timestamp');
        $microtimeCurrentState = (float) $this->state->get('timestamp');
        if ($microtimeUpdate > $microtimeCurrentState) {
            $this->state->modify($updates->all());
            $this->trigger('updates', $updates);
        }
        
        return $message->acknowlege();
    }

}
