<?php

namespace InSegment\WorkersFramework\Services;

use React\Promise;

use InSegment\WorkersFramework\Interfaces\QueueMessageInterface;
use InSegment\WorkersFramework\Interfaces\QueueDriverInterface;
use InSegment\WorkersFramework\Interfaces\ConsumerInterface;
use InSegment\WorkersFramework\Traits\EventsTrait;
use InSegment\WorkersFramework\Models\JobHeaders;
use InSegment\WorkersFramework\Models\Config;

class WorkerConsumer implements ConsumerInterface
{
    use EventsTrait;
    
    /**
     * Instance of queue driver connected to the channel
     *
     * @var \InSegment\WorkersFramework\Interfaces\QueueDriverInterface
     */
    protected $queueDriver;
    
    /**
     * Config instance
     *
     * @var \InSegment\WorkersFramework\Models\Config
     */
    protected $config;
    
    /**
     * Whether the consumer is started
     *
     * @var bool
     */
    protected $active;
    
    /**
     * Tag to track active subscription
     *
     * @var string|null
     */
    protected $consumerTag;
    
    /**
     * Name of the worker queue
     *
     * @var string
     */
    protected $queueName;
    
    /**
     * Constructor
     * 
     * @param \InSegment\WorkersFramework\Interfaces\QueueDriverInterface $driver
     * @param \InSegment\WorkersFramework\Models\Config $config
     * @param string $queueName
     */
    public function __construct(QueueDriverInterface $driver, Config $config, string $queueName)
    {
        $this->queueDriver = $driver;
        $this->config = $config;
        $this->queueName = $queueName;
    }
    
    /**
     * Activate consumption of the main queue
     * 
     * @return \React\Promise\PromiseInterface
     */
    public function activate(): Promise\PromiseInterface
    {
        echo "Worker consumer starting: {$this->queueName}\n";
        
        if ($this->active) { //  already started
            echo "already started: {$this->queueName}\n";
            return Promise\resolve();
        }

        return $this->queueDriver->subscribe($this->queueName, $this)->then(function ($consumerTag) {
            $this->active = true;
            $this->consumerTag = $consumerTag;
            echo "Worker consumer started: {$this->queueName}\n";
        });
    }
    
    /**
     * Deactivate consumption of the main queue
     * 
     * @return \React\Promise\PromiseInterface
     */
    public function deactivate(): Promise\PromiseInterface
    {
        echo "Worker consumer stopping: {$this->queueName}\n";
        
        if (!$this->active) {
            echo "not started: {$this->queueName}\n";
            return Promise\resolve();
        }
        
        return $this->queueDriver->unsubscribe($this->consumerTag)->then(function () {
            $this->active = false;
        });
    }
    
    /**
     * Move message from the main queue to the worker queue and acknowledge the original one
     * 
     * @param \InSegment\WorkersFramework\Interfaces\QueueMessageInterface $message
     * @param \InSegment\WorkersFramework\Interfaces\QueueDriverInterface $driver
     * @return mixed
     */
    public function receiveMessage(QueueMessageInterface $message, QueueDriverInterface $driver)
    {
        $headers = JobHeaders::extend($message->getHeaders());
        $headersValidation = $headers->validate();
        
        if (!$headersValidation->isValid()) {
            $printNotFoundFields = implode(', ', $headersValidation->getInvalidFields());
            echo "Job header mandatory fields not found: {$printNotFoundFields}, message will be ignored\n";
            return $message->acknowlege();
        }
        
        $jobName = $headers->getJobName();
        $job = $this->config->getJob($jobName);
        
        if ($job === null) {
            echo "The configuration doesn't have class mapping for job named {$jobName}, message will be ignored\n";
            return $message->acknowlege();
        }
        
        $this->trigger('job_start', $message);
        return Promise\resolve($job->process($message))->then(function ($result) use ($message) {
            $this->trigger('job_end', $message, $result);
            
            if ($result !== false) {
                return $message->acknowlege();
            }
            
            return $result;
        });
    }

}
